from calliope._version import __version__
from calliope.core import AttrDict, Model, read_netcdf, _logger, set_log_level
from calliope import examples, exceptions


__title__ = 'Calliope'
__author__ = 'Calliope contributors listed in AUTHORS'
__copyright__ = 'Copyright (C) 2013-2018 Calliope contributors listed in AUTHORS'
